@@ -61,6 +61,7 @@ struct erofs_buffer_head;
 struct erofs_bufmgr;
 
 struct erofs_device_info {
+	char *src_path;
 	u8 tag[64];
 	u32 blocks;
 	u32 mapped_blkaddr;
@@ -82,6 +83,7 @@ struct erofs_xattr_prefix_item {
 
 struct erofs_mkfs_dfops;
 struct erofs_packed_inode;
+struct z_erofs_mgr;
 
 struct erofs_sb_info {
 	struct erofs_sb_lz4_info lz4;
@@ -141,7 +143,9 @@ struct erofs_sb_info {
 	struct erofs_mkfs_dfops *mkfs_dfops;
 #endif
 	struct erofs_bufmgr *bmgr;
+	struct z_erofs_mgr *zmgr;
 	struct erofs_packed_inode *packedinode;
+	struct erofs_buffer_head *bh_devt;
 	bool useqpl;
 };
 
@@ -272,6 +276,7 @@ struct erofs_inode {
 			union {
 				unsigned int z_idataoff;
 				erofs_off_t fragmentoff;
+				void *fragment;
 			};
 #define z_idata_size	idata_size
 		};
@@ -371,10 +376,12 @@ enum {
 /* The length of extent is full */
 #define EROFS_MAP_FULL_MAPPED	(1 << BH_FullMapped)
 /* Located in the special packed inode */
-#define EROFS_MAP_FRAGMENT	(1 << BH_Fragment)
+#define __EROFS_MAP_FRAGMENT	(1 << BH_Fragment)
 /* The extent refers to partial decompressed data */
 #define EROFS_MAP_PARTIAL_REF	(1 << BH_Partialref)
 
+#define EROFS_MAP_FRAGMENT	(EROFS_MAP_MAPPED | __EROFS_MAP_FRAGMENT)
+
 struct erofs_map_blocks {
 	char mpage[EROFS_MAX_BLOCK_SIZE];
 
@@ -409,10 +416,12 @@ struct erofs_map_dev {
 /* super.c */
 int erofs_read_superblock(struct erofs_sb_info *sbi);
 void erofs_put_super(struct erofs_sb_info *sbi);
-int erofs_writesb(struct erofs_sb_info *sbi, struct erofs_buffer_head *sb_bh,
-		  erofs_blk_t *blocks);
+int erofs_writesb(struct erofs_sb_info *sbi, struct erofs_buffer_head *sb_bh);
 struct erofs_buffer_head *erofs_reserve_sb(struct erofs_bufmgr *bmgr);
+int erofs_mkfs_init_devices(struct erofs_sb_info *sbi, unsigned int devices);
+int erofs_write_device_table(struct erofs_sb_info *sbi);
 int erofs_enable_sb_chksum(struct erofs_sb_info *sbi, u32 *crc);
+int erofs_superblock_csum_verify(struct erofs_sb_info *sbi);
 
 /* namei.c */
 int erofs_read_inode_from_disk(struct erofs_inode *vi);
@@ -505,6 +514,9 @@ static inline int erofs_blk_read(struct erofs_sb_info *sbi, int device_id,
 			      erofs_pos(sbi, nblocks));
 }
 
+/* vmdk.c */
+int erofs_dump_vmdk_desc(FILE *f, struct erofs_sb_info *sbi);
+
 #ifdef EUCLEAN
 #define EFSCORRUPTED	EUCLEAN		/* Filesystem is corrupted */
 #else
