@@ -28,13 +28,6 @@
 
 #define Z_EROFS_DESTBUF_SZ	(Z_EROFS_PCLUSTER_MAX_SIZE + EROFS_MAX_BLOCK_SIZE * 2)
 
-/* compressing configuration specified by users */
-struct erofs_compress_cfg {
-	struct erofs_compress handle;
-	unsigned int algorithmtype;
-	bool enable;
-} erofs_ccfg[EROFS_MAX_COMPR_CFGS];
-
 struct z_erofs_extent_item {
 	struct list_head list;
 	struct z_erofs_inmem_extent e;
@@ -46,7 +39,7 @@ struct z_erofs_compress_ictx {		/* inode context */
 	int fd;
 	u64 fpos;
 
-	u32 tof_chksum;
+	u32 tofh;
 	bool fix_dedupedfrag;
 	bool fragemitted;
 	bool dedupe;
@@ -66,7 +59,10 @@ struct z_erofs_compress_ictx {		/* inode context */
 };
 
 struct z_erofs_compress_sctx {		/* segment context */
-	struct list_head extents;
+	union {
+		struct list_head extents;
+		struct list_head sibling;
+	};
 	struct z_erofs_compress_ictx *ictx;
 
 	u8 *queue;
@@ -111,13 +107,34 @@ struct erofs_compress_work {
 };
 
 static struct {
-	struct erofs_workqueue wq, fwq;
+	struct erofs_workqueue wq;
 	struct erofs_compress_work *idle;
 	pthread_mutex_t mutex;
 	bool hasfwq;
 } z_erofs_mt_ctrl;
+
+struct z_erofs_compress_fslot {
+	struct list_head pending;
+	pthread_mutex_t lock;
+	bool inprogress;
+};
+
 #endif
 
+/* compressing configuration specified by users */
+struct erofs_compress_cfg {
+	struct erofs_compress handle;
+	unsigned int algorithmtype;
+	bool enable;
+};
+
+struct z_erofs_mgr {
+	struct erofs_compress_cfg ccfg[EROFS_MAX_COMPR_CFGS];
+#ifdef EROFS_MT_ENABLED
+	struct z_erofs_compress_fslot fslot[1024];
+#endif
+};
+
 static bool z_erofs_mt_enabled;
 
 #define Z_EROFS_LEGACY_MAP_HEADER_SIZE	Z_EROFS_FULL_INDEX_ALIGN(0)
@@ -386,7 +403,7 @@ static int write_uncompressed_block(struct z_erofs_compress_sctx *ctx,
 	memcpy(dst, ctx->queue + ctx->head + rightpart, count - rightpart);
 
 	if (ctx->membuf) {
-		erofs_dbg("Writing %u uncompressed data of %s", count,
+		erofs_dbg("Recording %u uncompressed data of %s", count,
 			  inode->i_srcpath);
 		memcpy(ctx->membuf + ctx->memoff, dst, erofs_blksiz(sbi));
 		ctx->memoff += erofs_blksiz(sbi);
@@ -526,11 +543,8 @@ static bool z_erofs_fixup_deduped_fragment(struct z_erofs_compress_sctx *ctx)
 		return false;
 	}
 
-	inode->fragmentoff += inode->fragment_size - newsize;
 	inode->fragment_size = newsize;
-
-	erofs_dbg("Reducing fragment size to %llu at %llu",
-		  inode->fragment_size | 0ULL, inode->fragmentoff | 0ULL);
+	erofs_dbg("Reducing fragment size to %llu", inode->fragment_size | 0ULL);
 
 	/* it's the end */
 	DBG_BUGON(ctx->tail - ctx->head + ctx->remaining != newsize);
@@ -625,8 +639,8 @@ nocompression:
 		   compressedsize < ctx->pclustersize &&
 		   (!inode->fragment_size || ictx->fix_dedupedfrag)) {
 frag_packing:
-		ret = z_erofs_pack_fragments(inode, ctx->queue + ctx->head,
-					     len, ictx->tof_chksum);
+		ret = erofs_fragment_pack(inode, ctx->queue + ctx->head,
+					  ~0ULL, len, ictx->tofh, false);
 		if (ret < 0)
 			return ret;
 		e->plen = 0;	/* indicate a fragment */
@@ -691,7 +705,7 @@ frag_packing:
 
 		/* write compressed data */
 		if (ctx->membuf) {
-			erofs_dbg("Writing %u compressed data of %u bytes of %s",
+			erofs_dbg("Recording %u compressed data of %u bytes of %s",
 				  e->length, e->plen, inode->i_srcpath);
 
 			memcpy(ctx->membuf + ctx->memoff,
@@ -1103,7 +1117,7 @@ int z_erofs_compress_segment(struct z_erofs_compress_sctx *ctx,
 	DBG_BUGON(offset != -1 && frag && inode->fragment_size);
 	if (offset != -1 && frag && !inode->fragment_size &&
 	    cfg.c_fragdedupe != FRAGDEDUPE_OFF) {
-		ret = z_erofs_fragments_dedupe(inode, fd, &ictx->tof_chksum);
+		ret = erofs_fragment_findmatch(inode, fd, ictx->tofh);
 		if (ret < 0)
 			return ret;
 		if (inode->fragment_size > ctx->remaining)
@@ -1172,6 +1186,9 @@ int erofs_commit_compressed_file(struct z_erofs_compress_ictx *ictx,
 	int ret;
 
 	if (inode->fragment_size) {
+		ret = erofs_fragment_commit(inode, ictx->tofh);
+		if (ret)
+			goto err_free_idata;
 		inode->z_advise |= Z_EROFS_ADVISE_FRAGMENT_PCLUSTER;
 		erofs_sb_set_fragments(inode->sbi);
 	}
@@ -1190,6 +1207,7 @@ int erofs_commit_compressed_file(struct z_erofs_compress_ictx *ictx,
 	/* estimate if data compression saves space or not */
 	if (!inode->fragment_size && ptotal + inode->idata_size +
 	    legacymetasize >= inode->i_size) {
+		z_erofs_dedupe_ext_commit(true);
 		z_erofs_dedupe_commit(true);
 		ret = -ENOSPC;
 		goto err_free_meta;
@@ -1210,12 +1228,10 @@ int erofs_commit_compressed_file(struct z_erofs_compress_ictx *ictx,
 		legacymetasize = Z_EROFS_LEGACY_MAP_HEADER_SIZE;
 	}
 
-	if (ptotal) {
+	if (ptotal)
 		(void)erofs_bh_balloon(bh, ptotal);
-	} else {
-		if (!cfg.c_fragments && !cfg.c_dedupe)
-			DBG_BUGON(!inode->idata_size);
-	}
+	else if (!cfg.c_fragments && !cfg.c_dedupe)
+		DBG_BUGON(!inode->idata_size);
 
 	erofs_info("compressed %s (%llu bytes) into %llu bytes",
 		   inode->i_srcpath, inode->i_size | 0ULL, ptotal | 0ULL);
@@ -1238,9 +1254,6 @@ int erofs_commit_compressed_file(struct z_erofs_compress_ictx *ictx,
 		DBG_BUGON(ret);
 	}
 	inode->compressmeta = compressmeta;
-	if (!erofs_is_packed_inode(inode))
-		erofs_droid_blocklist_write(inode, pstart >> bbits,
-					    inode->u.i_blocks);
 	return 0;
 
 err_free_meta:
@@ -1365,6 +1378,32 @@ out:
 	pthread_mutex_unlock(&ictx->mutex);
 }
 
+void z_erofs_mt_f_workfn(struct erofs_work *work, void *tlsp)
+{
+	struct erofs_compress_work *cwork = (struct erofs_compress_work *)work;
+	struct erofs_sb_info *sbi = cwork->ctx.ictx->inode->sbi;
+	u32 tofh = cwork->ctx.ictx->tofh;
+	struct z_erofs_compress_fslot *fs = &sbi->zmgr->fslot[tofh & 1023];
+
+	while (1) {
+		z_erofs_mt_workfn(work, tlsp);
+		pthread_mutex_lock(&fs->lock);
+
+		if (list_empty(&fs->pending)) {
+			fs->inprogress = false;
+			pthread_mutex_unlock(&fs->lock);
+			break;
+		}
+		cwork = list_first_entry(&fs->pending,
+					 struct erofs_compress_work,
+					 ctx.sibling);
+		list_del(&cwork->ctx.sibling);
+		pthread_mutex_unlock(&fs->lock);
+		init_list_head(&cwork->ctx.extents);
+		work = &cwork->work;
+	}
+}
+
 int z_erofs_merge_segment(struct z_erofs_compress_ictx *ictx,
 			  struct z_erofs_compress_sctx *sctx)
 {
@@ -1373,7 +1412,7 @@ int z_erofs_merge_segment(struct z_erofs_compress_ictx *ictx,
 	bool dedupe_ext = cfg.c_fragments;
 	erofs_off_t off = 0;
 	int ret = 0, ret2;
-	erofs_blk_t dupb;
+	erofs_off_t dpo;
 	u64 hash;
 
 	list_for_each_entry_safe(ei, n, &sctx->extents, list) {
@@ -1391,10 +1430,10 @@ int z_erofs_merge_segment(struct z_erofs_compress_ictx *ictx,
 			continue;
 
 		if (dedupe_ext) {
-			dupb = z_erofs_dedupe_ext_match(sbi, sctx->membuf + off,
+			dpo = z_erofs_dedupe_ext_match(sbi, sctx->membuf + off,
 						ei->e.plen, ei->e.raw, &hash);
-			if (dupb != EROFS_NULL_ADDR) {
-				ei->e.pstart = dupb;
+			if (dpo) {
+				ei->e.pstart = dpo;
 				sctx->pstart -= ei->e.plen;
 				off += ei->e.plen;
 				ictx->dedupe = true;
@@ -1406,6 +1445,8 @@ int z_erofs_merge_segment(struct z_erofs_compress_ictx *ictx,
 				continue;
 			}
 		}
+		erofs_dbg("Writing %u %scompressed data of %s to %llu", ei->e.length,
+			  ei->e.raw ? "un" : "", ictx->inode->i_srcpath, ei->e.pstart);
 		ret2 = erofs_dev_write(sbi, sctx->membuf + off, ei->e.pstart,
 				       ei->e.plen);
 		off += ei->e.plen;
@@ -1464,18 +1505,32 @@ int z_erofs_mt_compress(struct z_erofs_compress_ictx *ictx)
 		cur->dict_size = ccfg->handle.dict_size;
 		cur->errcode = 1;	/* mark as "in progress" */
 
-		cur->work.fn = z_erofs_mt_workfn;
 		if (i >= nsegs - 1) {
 			cur->ctx.remaining = inode->i_size -
 					inode->fragment_size - (u64)i * segsz;
-			if (z_erofs_mt_ctrl.hasfwq) {
-				erofs_queue_work(&z_erofs_mt_ctrl.fwq,
+
+			if (z_erofs_mt_ctrl.hasfwq && ictx->tofh != ~0U) {
+				struct z_erofs_mgr *zmgr = inode->sbi->zmgr;
+				struct z_erofs_compress_fslot *fs =
+					&zmgr->fslot[ictx->tofh & 1023];
+
+				pthread_mutex_lock(&fs->lock);
+				if (fs->inprogress) {
+					list_add_tail(&cur->ctx.sibling,
+						      &fs->pending);
+				} else {
+					fs->inprogress = true;
+					cur->work.fn = z_erofs_mt_f_workfn;
+					erofs_queue_work(&z_erofs_mt_ctrl.wq,
 						 &cur->work);
+				}
+				pthread_mutex_unlock(&fs->lock);
 				continue;
 			}
 		} else {
 			cur->ctx.remaining = segsz;
 		}
+		cur->work.fn = z_erofs_mt_workfn;
 		erofs_queue_work(&z_erofs_mt_ctrl.wq, &cur->work);
 	}
 	ictx->mtworks = head;
@@ -1532,7 +1587,6 @@ int erofs_mt_write_compressed_file(struct z_erofs_compress_ictx *ictx)
 	ret = erofs_commit_compressed_file(ictx, bh, pstart - ptotal, ptotal);
 
 out:
-	close(ictx->fd);
 	free(ictx);
 	return ret;
 }
@@ -1548,15 +1602,8 @@ static int z_erofs_mt_init(void)
 		erofs_warn("multi-threaded dedupe is NOT implemented for now");
 		cfg.c_mt_workers = 0;
 	} else {
-		if (cfg.c_fragments && workers > 1) {
-			ret = erofs_alloc_workqueue(&z_erofs_mt_ctrl.fwq, 1, 32,
-						    z_erofs_mt_wq_tls_alloc,
-						    z_erofs_mt_wq_tls_free);
-			if (ret)
-				return ret;
+		if (cfg.c_fragments && workers > 1)
 			z_erofs_mt_ctrl.hasfwq = true;
-			--workers;
-		}
 
 		ret = erofs_alloc_workqueue(&z_erofs_mt_ctrl.wq, workers,
 					    workers << 2,
@@ -1596,7 +1643,7 @@ void *erofs_begin_compressed_file(struct erofs_inode *inode, int fd, u64 fpos)
 		while (1) {
 			inode->z_algorithmtype[0] =
 				rand() % EROFS_MAX_COMPR_CFGS;
-			if (erofs_ccfg[inode->z_algorithmtype[0]].enable)
+			if (sbi->zmgr->ccfg[inode->z_algorithmtype[0]].enable)
 				break;
 		}
 	}
@@ -1613,33 +1660,34 @@ void *erofs_begin_compressed_file(struct erofs_inode *inode, int fd, u64 fpos)
 		pthread_mutex_unlock(&g_ictx.mutex);
 #endif
 		ictx = &g_ictx;
-		ictx->fd = fd;
 	} else {
 		ictx = malloc(sizeof(*ictx));
 		if (!ictx)
 			return ERR_PTR(-ENOMEM);
-		ictx->fd = dup(fd);
 	}
+	ictx->fd = fd;
 
-	ictx->ccfg = &erofs_ccfg[inode->z_algorithmtype[0]];
+	ictx->ccfg = &sbi->zmgr->ccfg[inode->z_algorithmtype[0]];
 	inode->z_algorithmtype[0] = ictx->ccfg->algorithmtype;
 	inode->z_algorithmtype[1] = 0;
 
-	/*
-	 * Handle tails in advance to avoid writing duplicated
-	 * parts into the packed inode.
-	 */
-	if (cfg.c_fragments && !erofs_is_packed_inode(inode) &&
-	    ictx == &g_ictx && cfg.c_fragdedupe != FRAGDEDUPE_OFF) {
-		ret = z_erofs_fragments_dedupe(inode, fd, &ictx->tof_chksum);
-		if (ret < 0)
-			goto err_free_ictx;
+	if (cfg.c_fragments && !erofs_is_packed_inode(inode)) {
+		ictx->tofh = z_erofs_fragments_tofh(inode, fd, fpos);
+		if (ictx == &g_ictx && cfg.c_fragdedupe != FRAGDEDUPE_OFF) {
+			/*
+			 * Handle tails in advance to avoid writing duplicated
+			 * parts into the packed inode.
+			 */
+			ret = erofs_fragment_findmatch(inode, fd, ictx->tofh);
+			if (ret < 0)
+				goto err_free_ictx;
 
-		if (cfg.c_fragdedupe == FRAGDEDUPE_INODE &&
-		    inode->fragment_size < inode->i_size) {
-			erofs_dbg("Discard the sub-inode tail fragment of %s",
-				  inode->i_srcpath);
-			inode->fragment_size = 0;
+			if (cfg.c_fragdedupe == FRAGDEDUPE_INODE &&
+			    inode->fragment_size < inode->i_size) {
+				erofs_dbg("Discard the sub-inode tail fragment of %s",
+					  inode->i_srcpath);
+				inode->fragment_size = 0;
+			}
 		}
 	}
 	ictx->inode = inode;
@@ -1650,7 +1698,7 @@ void *erofs_begin_compressed_file(struct erofs_inode *inode, int fd, u64 fpos)
 	ictx->dedupe = false;
 
 	if (all_fragments && !inode->fragment_size) {
-		ret = z_erofs_pack_file_from_fd(inode, fd, ictx->tof_chksum);
+		ret = erofs_pack_file_from_fd(inode, fd, ictx->tofh);
 		if (ret)
 			goto err_free_idata;
 	}
@@ -1847,8 +1895,15 @@ int z_erofs_compress_init(struct erofs_sb_info *sbi, struct erofs_buffer_head *s
 	u32 max_dict_size[Z_EROFS_COMPRESSION_MAX] = {};
 	u32 available_compr_algs = 0;
 
+	if (!sbi->zmgr) {
+		sbi->zmgr = calloc(1, sizeof(*sbi->zmgr));
+		if (!sbi->zmgr)
+			return -ENOMEM;
+	}
+
 	for (i = 0; cfg.c_compr_opts[i].alg; ++i) {
-		struct erofs_compress *c = &erofs_ccfg[i].handle;
+		struct erofs_compress_cfg *ccfg = &sbi->zmgr->ccfg[i];
+		struct erofs_compress *c = &ccfg->handle;
 
 		ret = erofs_compressor_init(sbi, c, cfg.c_compr_opts[i].alg,
 					    cfg.c_compr_opts[i].level,
@@ -1857,10 +1912,10 @@ int z_erofs_compress_init(struct erofs_sb_info *sbi, struct erofs_buffer_head *s
 			return ret;
 
 		id = z_erofs_get_compress_algorithm_id(c);
-		erofs_ccfg[i].algorithmtype = id;
-		erofs_ccfg[i].enable = true;
-		available_compr_algs |= 1 << erofs_ccfg[i].algorithmtype;
-		if (erofs_ccfg[i].algorithmtype != Z_EROFS_COMPRESSION_LZ4)
+		ccfg->algorithmtype = id;
+		ccfg->enable = true;
+		available_compr_algs |= 1 << ccfg->algorithmtype;
+		if (ccfg->algorithmtype != Z_EROFS_COMPRESSION_LZ4)
 			erofs_sb_set_compr_cfgs(sbi);
 		if (c->dict_size > max_dict_size[id])
 			max_dict_size[id] = c->dict_size;
@@ -1921,15 +1976,31 @@ int z_erofs_compress_init(struct erofs_sb_info *sbi, struct erofs_buffer_head *s
 	}
 
 	z_erofs_mt_enabled = false;
-	return z_erofs_mt_init();
+	ret = z_erofs_mt_init();
+	if (ret)
+		return ret;
+
+#ifdef EROFS_MT_ENABLED
+	if (z_erofs_mt_ctrl.hasfwq) {
+		for (i = 0; i < ARRAY_SIZE(sbi->zmgr->fslot); ++i) {
+			init_list_head(&sbi->zmgr->fslot[i].pending);
+			pthread_mutex_init(&sbi->zmgr->fslot[i].lock, NULL);
+		}
+	}
+#endif
+	return 0;
 }
 
-int z_erofs_compress_exit(void)
+int z_erofs_compress_exit(struct erofs_sb_info *sbi)
 {
 	int i, ret;
 
+	/* If `zmgr` is uninitialized, return directly. */
+	if (!sbi->zmgr)
+		return 0;
+
 	for (i = 0; cfg.c_compr_opts[i].alg; ++i) {
-		ret = erofs_compressor_exit(&erofs_ccfg[i].handle);
+		ret = erofs_compressor_exit(&sbi->zmgr->ccfg[i].handle);
 		if (ret)
 			return ret;
 	}
@@ -1937,9 +2008,6 @@ int z_erofs_compress_exit(void)
 	if (z_erofs_mt_enabled) {
 #ifdef EROFS_MT_ENABLED
 		ret = erofs_destroy_workqueue(&z_erofs_mt_ctrl.wq);
-		if (ret)
-			return ret;
-		ret = erofs_destroy_workqueue(&z_erofs_mt_ctrl.fwq);
 		if (ret)
 			return ret;
 		while (z_erofs_mt_ctrl.idle) {
