@@ -100,7 +100,7 @@ int __erofs_map_blocks(struct erofs_inode *inode,
 		return -EIO;
 
 	map->m_la = chunknr << vi->u.chunkbits;
-	map->m_plen = min_t(erofs_off_t, 1UL << vi->u.chunkbits,
+	map->m_plen = min_t(erofs_off_t, 1ULL << vi->u.chunkbits,
 			roundup(inode->i_size - map->m_la, erofs_blksiz(sbi)));
 
 	/* handle block map */
@@ -248,7 +248,12 @@ int z_erofs_read_one_data(struct erofs_inode *inode,
 	struct erofs_map_dev mdev;
 	int ret = 0;
 
-	if (map->m_flags & EROFS_MAP_FRAGMENT) {
+	if (map->m_flags & __EROFS_MAP_FRAGMENT) {
+		if (__erofs_unlikely(inode->nid == sbi->packed_nid)) {
+			erofs_err("fragment should not exist in the packed inode %llu",
+				  sbi->packed_nid | 0ULL);
+			return -EFSCORRUPTED;
+		}
 		return erofs_packedfile_read(sbi, buffer, length - skip,
 				   inode->fragmentoff + skip);
 	}
