@@ -9,6 +9,8 @@
 #include "erofs/err.h"
 #include "erofs/print.h"
 
+#if defined(HAVE_LIBZSTD) || defined(HAVE_QPL) || defined(HAVE_LIBDEFLATE) || \
+    defined(HAVE_ZLIB) || defined(HAVE_LIBLZMA) || defined(LZ4_ENABLED)
 static unsigned int z_erofs_fixup_insize(const u8 *padbuf, unsigned int padbufsize)
 {
 	unsigned int inputmargin;
@@ -17,6 +19,7 @@ static unsigned int z_erofs_fixup_insize(const u8 *padbuf, unsigned int padbufsi
 	     !padbuf[inputmargin]; ++inputmargin);
 	return inputmargin;
 }
+#endif
 
 #ifdef HAVE_LIBZSTD
 #include <zstd.h>
@@ -511,16 +514,15 @@ int z_erofs_decompress(struct z_erofs_decompress_req *rq)
 	if (rq->alg == Z_EROFS_COMPRESSION_INTERLACED) {
 		unsigned int count, rightpart, skip;
 
+		if (rq->decodedlength > rq->inputsize)
+			return -EOPNOTSUPP;
+		if (rq->decodedlength < rq->decodedskip)
+			return -EFSCORRUPTED;
+
 		/* XXX: should support inputsize >= erofs_blksiz(sbi) later */
 		if (rq->inputsize > erofs_blksiz(sbi))
 			return -EFSCORRUPTED;
 
-		if (rq->decodedlength > erofs_blksiz(sbi))
-			return -EFSCORRUPTED;
-
-		if (rq->decodedlength < rq->decodedskip)
-			return -EFSCORRUPTED;
-
 		count = rq->decodedlength - rq->decodedskip;
 		skip = erofs_blkoff(sbi, rq->interlaced_offset + rq->decodedskip);
 		rightpart = min(erofs_blksiz(sbi) - skip, count);
@@ -529,9 +531,10 @@ int z_erofs_decompress(struct z_erofs_decompress_req *rq)
 		return 0;
 	} else if (rq->alg == Z_EROFS_COMPRESSION_SHIFTED) {
 		if (rq->decodedlength > rq->inputsize)
+			return -EOPNOTSUPP;
+		if (rq->decodedlength < rq->decodedskip)
 			return -EFSCORRUPTED;
 
-		DBG_BUGON(rq->decodedlength < rq->decodedskip);
 		memcpy(rq->out, rq->in + rq->decodedskip,
 		       rq->decodedlength - rq->decodedskip);
 		return 0;
