@@ -63,7 +63,6 @@ static struct option long_options[] = {
 #ifdef WITH_ANDROID
 	{"product-out", required_argument, NULL, 513},
 	{"fs-config-file", required_argument, NULL, 514},
-	{"block-list-file", required_argument, NULL, 515},
 #endif
 	{"ovlfs-strip", optional_argument, NULL, 516},
 	{"offset", required_argument, NULL, 517},
@@ -87,6 +86,11 @@ static struct option long_options[] = {
 	{"sort", required_argument, NULL, 527},
 	{"hard-dereference", no_argument, NULL, 528},
 	{"dsunit", required_argument, NULL, 529},
+#ifdef EROFS_MT_ENABLED
+	{"async-queue-limit", required_argument, NULL, 530},
+#endif
+	{"fsalignblks", required_argument, NULL, 531},
+	{"vmdk-desc", required_argument, NULL, 532},
 	{0, 0, 0, 0},
 };
 
@@ -160,6 +164,9 @@ static void usage(int argc, char **argv)
 		"    --mkfs-time        the timestamp is applied as build time only\n"
 		" -UX                   use a given filesystem UUID\n"
 		" --all-root            make all files owned by root\n"
+#ifdef EROFS_MT_ENABLED
+		" --async-queue-limit=# specify the maximum number of entries in the multi-threaded job queue\n"
+#endif
 		" --blobdev=X           specify an extra device X to store chunked data\n"
 		" --chunksize=#         generate chunk-based files with #-byte chunks\n"
 		" --clean=X             run full clean build (default) or:\n"
@@ -175,6 +182,7 @@ static void usage(int argc, char **argv)
 #endif
 		" --force-uid=#         set all file uids to # (# = UID)\n"
 		" --force-gid=#         set all file gids to # (# = GID)\n"
+		" --fsalignblks=#       specify the alignment of the primary device size in blocks\n"
 		" --uid-offset=#        add offset # to all file uids (# = id offset)\n"
 		" --gid-offset=#        add offset # to all file gids (# = id offset)\n"
 		" --hard-dereference    dereference hardlinks, add links as separate inodes\n"
@@ -203,6 +211,7 @@ static void usage(int argc, char **argv)
 		" --unxz[=X]            try to filter the tarball stream through xz/lzma/lzip\n"
 		"                       (and optionally dump the raw stream to X together)\n"
 #endif
+		" --vmdk-desc=X         generate a VMDK descriptor file to merge sub-filesystems\n"
 #ifdef EROFS_MT_ENABLED
 		" --workers=#           set the number of worker threads to # (default: %u)\n"
 #endif
@@ -213,7 +222,6 @@ static void usage(int argc, char **argv)
 		"Android-specific options:\n"
 		" --product-out=X       X=product_out directory\n"
 		" --fs-config-file=X    X=fs_config file\n"
-		" --block-list-file=X   X=block_list file\n"
 #endif
 #ifdef EROFS_MT_ENABLED
 		, erofs_get_available_processors() /* --workers= */
@@ -246,6 +254,9 @@ static LIST_HEAD(rebuild_src_list);
 static u8 fixeduuid[16];
 static bool valid_fixeduuid;
 static unsigned int dsunit;
+static unsigned int fsalignblks = 1;
+static int tarerofs_decoder;
+static FILE *vmdk_dcf;
 
 static int erofs_mkfs_feat_set_legacy_compress(bool en, const char *val,
 					       unsigned int vallen)
@@ -570,12 +581,93 @@ static void erofs_rebuild_cleanup(void)
 	rebuild_src_count = 0;
 }
 
+static int mkfs_parse_sources(int argc, char *argv[], int optind)
+{
+	struct stat st;
+	int err, fd;
+	char *s;
+
+	if (tar_mode) {
+		cfg.c_src_path = strdup(argv[optind++]);
+		if (!cfg.c_src_path)
+			return -ENOMEM;
+		fd = open(cfg.c_src_path, O_RDONLY);
+		if (fd < 0) {
+			erofs_err("failed to open tar file: %s", cfg.c_src_path);
+			return -errno;
+		}
+		err = erofs_iostream_open(&erofstar.ios, fd,
+					  tarerofs_decoder);
+		if (err)
+			return err;
+
+		if (erofstar.dumpfile) {
+			fd = open(erofstar.dumpfile,
+				  O_WRONLY | O_CREAT | O_TRUNC, 0644);
+			if (fd < 0) {
+				erofs_err("failed to open dumpfile: %s",
+					  erofstar.dumpfile);
+				return -errno;
+			}
+			erofstar.ios.dumpfd = fd;
+		}
+	} else {
+		err = lstat((s = argv[optind++]), &st);
+		if (err) {
+			erofs_err("failed to stat %s: %s", s,
+				  erofs_strerror(-errno));
+			return -ENOENT;
+		}
+		if (S_ISDIR(st.st_mode)) {
+			cfg.c_src_path = realpath(s, NULL);
+			if (!cfg.c_src_path) {
+				erofs_err("failed to parse source directory: %s",
+					  erofs_strerror(-errno));
+				return -ENOENT;
+			}
+			erofs_set_fs_root(cfg.c_src_path);
+		} else {
+			cfg.c_src_path = strdup(s);
+			if (!cfg.c_src_path)
+				return -ENOMEM;
+			rebuild_mode = true;
+		}
+	}
+
+	if (rebuild_mode) {
+		char *srcpath = cfg.c_src_path;
+		struct erofs_sb_info *src;
+
+		do {
+			src = calloc(1, sizeof(struct erofs_sb_info));
+			if (!src) {
+				erofs_rebuild_cleanup();
+				return -ENOMEM;
+			}
+
+			err = erofs_dev_open(src, srcpath, O_RDONLY);
+			if (err) {
+				free(src);
+				erofs_rebuild_cleanup();
+				return err;
+			}
+
+			/* extra device index starts from 1 */
+			src->dev = ++rebuild_src_count;
+			list_add(&src->list, &rebuild_src_list);
+		} while (optind < argc && (srcpath = argv[optind++]));
+	} else if (optind < argc) {
+		erofs_err("unexpected argument: %s\n", argv[optind]);
+		return -EINVAL;
+	}
+	return 0;
+}
+
 static int mkfs_parse_options_cfg(int argc, char *argv[])
 {
 	char *endptr;
 	int opt, i, err;
 	bool quiet = false;
-	int tarerofs_decoder = 0;
 	bool has_timestamp = false;
 
 	while ((opt = getopt_long(argc, argv, "C:E:L:T:U:b:d:x:z:Vh",
@@ -723,9 +815,6 @@ static int mkfs_parse_options_cfg(int argc, char *argv[])
 		case 514:
 			cfg.fs_config_file = optarg;
 			break;
-		case 515:
-			cfg.block_list_file = optarg;
-			break;
 #endif
 		case 'C':
 			i = strtoull(optarg, &endptr, 0);
@@ -886,6 +975,29 @@ static int mkfs_parse_options_cfg(int argc, char *argv[])
 				return -EINVAL;
 			}
 			break;
+#ifdef EROFS_MT_ENABLED
+		case 530:
+			cfg.c_mt_async_queue_limit = strtoul(optarg, &endptr, 0);
+			if (*endptr != '\0') {
+				erofs_err("invalid async-queue-limit %s", optarg);
+				return -EINVAL;
+			}
+			break;
+#endif
+		case 531:
+			fsalignblks = strtoul(optarg, &endptr, 0);
+			if (*endptr != '\0') {
+				erofs_err("invalid fsalignblks %s", optarg);
+				return -EINVAL;
+			}
+			break;
+		case 532:
+			vmdk_dcf = fopen(optarg, "wb");
+			if (!vmdk_dcf) {
+				erofs_err("failed to open vmdk desc `%s`", optarg);
+				return -EINVAL;
+			}
+			break;
 		case 'V':
 			version();
 			exit(0);
@@ -919,93 +1031,28 @@ static int mkfs_parse_options_cfg(int argc, char *argv[])
 	if (!cfg.c_img_path)
 		return -ENOMEM;
 
-	if (optind >= argc) {
-		if (!tar_mode) {
-			erofs_err("missing argument: SOURCE(s)");
-			return -EINVAL;
-		} else {
-			int dupfd;
-
-			dupfd = dup(STDIN_FILENO);
-			if (dupfd < 0) {
-				erofs_err("failed to duplicate STDIN_FILENO: %s",
-					  strerror(errno));
-				return -errno;
-			}
-			err = erofs_iostream_open(&erofstar.ios, dupfd,
-						  tarerofs_decoder);
-			if (err)
-				return err;
-		}
+	if (optind < argc) {
+		err = mkfs_parse_sources(argc, argv, optind);
+		if (err)
+			return err;
+	} else if (!tar_mode) {
+		erofs_err("missing argument: SOURCE(s)");
+		return -EINVAL;
 	} else {
-		struct stat st;
+		int dupfd;
 
-		cfg.c_src_path = realpath(argv[optind++], NULL);
-		if (!cfg.c_src_path) {
-			erofs_err("failed to parse source directory: %s",
-				  erofs_strerror(-errno));
-			return -ENOENT;
-		}
-
-		if (tar_mode) {
-			int fd = open(cfg.c_src_path, O_RDONLY);
-
-			if (fd < 0) {
-				erofs_err("failed to open file: %s", cfg.c_src_path);
-				return -errno;
-			}
-			err = erofs_iostream_open(&erofstar.ios, fd,
-						  tarerofs_decoder);
-			if (err)
-				return err;
-
-			if (erofstar.dumpfile) {
-				fd = open(erofstar.dumpfile,
-					  O_WRONLY | O_CREAT | O_TRUNC, 0644);
-				if (fd < 0) {
-					erofs_err("failed to open dumpfile: %s",
-						  erofstar.dumpfile);
-					return -errno;
-				}
-				erofstar.ios.dumpfd = fd;
-			}
-		} else {
-			err = lstat(cfg.c_src_path, &st);
-			if (err)
-				return -errno;
-			if (S_ISDIR(st.st_mode))
-				erofs_set_fs_root(cfg.c_src_path);
-			else
-				rebuild_mode = true;
-		}
-
-		if (rebuild_mode) {
-			char *srcpath = cfg.c_src_path;
-			struct erofs_sb_info *src;
-
-			do {
-				src = calloc(1, sizeof(struct erofs_sb_info));
-				if (!src) {
-					erofs_rebuild_cleanup();
-					return -ENOMEM;
-				}
-
-				err = erofs_dev_open(src, srcpath, O_RDONLY);
-				if (err) {
-					free(src);
-					erofs_rebuild_cleanup();
-					return err;
-				}
-
-				/* extra device index starts from 1 */
-				src->dev = ++rebuild_src_count;
-				list_add(&src->list, &rebuild_src_list);
-			} while (optind < argc && (srcpath = argv[optind++]));
-		} else if (optind < argc) {
-			erofs_err("unexpected argument: %s\n", argv[optind]);
-			return -EINVAL;
+		dupfd = dup(STDIN_FILENO);
+		if (dupfd < 0) {
+			erofs_err("failed to duplicate STDIN_FILENO: %s",
+				  strerror(errno));
+			return -errno;
 		}
+		err = erofs_iostream_open(&erofstar.ios, dupfd,
+					  tarerofs_decoder);
+		if (err)
+			return err;
 	}
+
 	if (quiet) {
 		cfg.c_dbg_lvl = EROFS_ERR;
 		cfg.c_showprogress = false;
@@ -1077,11 +1124,6 @@ int parse_source_date_epoch(void)
 			  source_date_epoch);
 		return -EINVAL;
 	}
-
-	if (cfg.c_force_inodeversion != FORCE_INODE_EXTENDED)
-		erofs_info("SOURCE_DATE_EPOCH is set, forcely generate extended inodes instead");
-
-	cfg.c_force_inodeversion = FORCE_INODE_EXTENDED;
 	cfg.c_unix_timestamp = epoch;
 	cfg.c_timeinherit = TIMESTAMP_CLAMPING;
 	return 0;
@@ -1095,6 +1137,7 @@ void erofs_show_progs(int argc, char *argv[])
 
 static int erofs_mkfs_rebuild_load_trees(struct erofs_inode *root)
 {
+	struct erofs_device_info *devs;
 	struct erofs_sb_info *src;
 	unsigned int extra_devices = 0;
 	erofs_blk_t nblocks;
@@ -1143,20 +1186,22 @@ static int erofs_mkfs_rebuild_load_trees(struct erofs_inode *root)
 	if (ret)
 		return ret;
 
+	devs = g_sbi.devs;
 	list_for_each_entry(src, &rebuild_src_list, list) {
 		u8 *tag = NULL;
 
+		DBG_BUGON(src->dev < 1);
+		idx = src->dev - 1;
 		if (extra_devices) {
 			nblocks = src->devs[0].blocks;
 			tag = src->devs[0].tag;
 		} else {
 			nblocks = src->primarydevice_blocks;
+			devs[idx].src_path = strdup(src->devname);
 		}
-		DBG_BUGON(src->dev < 1);
-		idx = src->dev - 1;
-		g_sbi.devs[idx].blocks = nblocks;
+		devs[idx].blocks = nblocks;
 		if (tag && *tag)
-			memcpy(g_sbi.devs[idx].tag, tag, sizeof(g_sbi.devs[0].tag));
+			memcpy(devs[idx].tag, tag, sizeof(devs[0].tag));
 		else
 			/* convert UUID of the source image to a hex string */
 			sprintf((char *)g_sbi.devs[idx].tag,
@@ -1173,7 +1218,7 @@ static int erofs_mkfs_rebuild_load_trees(struct erofs_inode *root)
 	return 0;
 }
 
-static void erofs_mkfs_showsummaries(erofs_blk_t nblocks)
+static void erofs_mkfs_showsummaries(void)
 {
 	char uuid_str[37] = {};
 	char *incr = incremental_mode ? "new" : "total";
@@ -1184,23 +1229,24 @@ static void erofs_mkfs_showsummaries(erofs_blk_t nblocks)
 	erofs_uuid_unparse_lower(g_sbi.uuid, uuid_str);
 
 	fprintf(stdout, "------\nFilesystem UUID: %s\n"
-		"Filesystem total blocks: %u (of %u-byte blocks)\n"
+		"Filesystem total blocks: %llu (of %u-byte blocks)\n"
 		"Filesystem total inodes: %llu\n"
 		"Filesystem %s metadata blocks: %u\n"
 		"Filesystem %s deduplicated bytes (of source files): %llu\n",
-		uuid_str, nblocks, 1U << g_sbi.blkszbits, g_sbi.inos | 0ULL,
+		uuid_str, g_sbi.total_blocks | 0ULL, 1U << g_sbi.blkszbits,
+		g_sbi.inos | 0ULL,
 		incr, erofs_total_metablocks(g_sbi.bmgr),
 		incr, g_sbi.saved_by_deduplication | 0ULL);
 }
 
 int main(int argc, char **argv)
 {
-	int err = 0;
 	struct erofs_buffer_head *sb_bh;
 	struct erofs_inode *root = NULL;
-	erofs_blk_t nblocks = 0;
+	bool tar_index_512b = false;
 	struct timeval t;
 	FILE *blklst = NULL;
+	int err = 0;
 	u32 crc;
 
 	erofs_init_configure();
@@ -1241,14 +1287,6 @@ int main(int argc, char **argv)
 		erofs_err("failed to load fs config %s", cfg.fs_config_file);
 		return 1;
 	}
-
-	if (cfg.block_list_file) {
-		blklst = fopen(cfg.block_list_file, "w");
-		if (!blklst || erofs_blocklist_open(blklst, false)) {
-			erofs_err("failed to open %s", cfg.block_list_file);
-			return 1;
-		}
-	}
 #endif
 	erofs_show_config();
 	if (cfg.c_fragments || cfg.c_extra_ea_name_prefixes) {
@@ -1279,12 +1317,13 @@ int main(int argc, char **argv)
 				erofs_err("failed to open %s", erofstar.mapfile);
 				goto exit;
 			}
-		} else if (erofstar.index_mode) {
+		} else if (erofstar.index_mode && !erofstar.headeronly_mode) {
 			/*
 			 * If mapfile is unspecified for tarfs index mode,
 			 * 512-byte block size is enforced here.
 			 */
 			g_sbi.blkszbits = 9;
+			tar_index_512b = true;
 		}
 	}
 
@@ -1400,8 +1439,7 @@ int main(int argc, char **argv)
 			return 1;
 	}
 
-	if (((erofstar.index_mode && !erofstar.headeronly_mode) &&
-	    !erofstar.mapfile) || cfg.c_blobdev_path) {
+	if (tar_index_512b || cfg.c_blobdev_path) {
 		err = erofs_mkfs_init_devices(&g_sbi, 1);
 		if (err) {
 			erofs_err("failed to generate device table: %s",
@@ -1459,8 +1497,16 @@ int main(int argc, char **argv)
 		}
 	}
 
-	if (erofstar.index_mode && g_sbi.extra_devices && !erofstar.mapfile)
-		g_sbi.devs[0].blocks = BLK_ROUND_UP(&g_sbi, erofstar.offset);
+	if (tar_index_512b) {
+		if (!g_sbi.extra_devices) {
+			DBG_BUGON(1);
+		} else {
+			if (cfg.c_src_path)
+				g_sbi.devs[0].src_path = strdup(cfg.c_src_path);
+			g_sbi.devs[0].blocks =
+				BLK_ROUND_UP(&g_sbi, erofstar.offset);
+		}
+	}
 
 	if ((cfg.c_fragments || cfg.c_extra_ea_name_prefixes) &&
 	    erofs_sb_has_fragments(&g_sbi)) {
@@ -1476,6 +1522,12 @@ int main(int argc, char **argv)
 			goto exit;
 	}
 
+	g_sbi.primarydevice_blocks =
+		roundup(erofs_mapbh(g_sbi.bmgr, NULL), fsalignblks);
+	err = erofs_write_device_table(&g_sbi);
+	if (err)
+		goto exit;
+
 	/* flush all buffers except for the superblock */
 	err = erofs_bflush(g_sbi.bmgr, NULL);
 	if (err)
@@ -1485,7 +1537,7 @@ int main(int argc, char **argv)
 	erofs_iput(root);
 	root = NULL;
 
-	err = erofs_writesb(&g_sbi, sb_bh, &nblocks);
+	err = erofs_writesb(&g_sbi, sb_bh);
 	if (err)
 		goto exit;
 
@@ -1494,17 +1546,22 @@ int main(int argc, char **argv)
 	if (err)
 		goto exit;
 
-	err = erofs_dev_resize(&g_sbi, nblocks);
+	err = erofs_dev_resize(&g_sbi, g_sbi.primarydevice_blocks);
 
 	if (!err && erofs_sb_has_sb_chksum(&g_sbi)) {
 		err = erofs_enable_sb_chksum(&g_sbi, &crc);
 		if (!err)
 			erofs_info("superblock checksum 0x%08x written", crc);
 	}
+
+	if (!err && vmdk_dcf) {
+		err = erofs_dump_vmdk_desc(vmdk_dcf, &g_sbi);
+		fclose(vmdk_dcf);
+	}
 exit:
 	if (root)
 		erofs_iput(root);
-	z_erofs_compress_exit();
+	z_erofs_compress_exit(&g_sbi);
 	z_erofs_dedupe_exit();
 	z_erofs_dedupe_ext_exit();
 	blklst = erofs_blocklist_close();
@@ -1532,7 +1589,7 @@ exit:
 		return 1;
 	}
 	erofs_update_progressinfo("Build completed.\n");
-	erofs_mkfs_showsummaries(nblocks);
+	erofs_mkfs_showsummaries();
 	erofs_put_super(&g_sbi);
 	return 0;
 }
